//////////////////////////////////////////
// Driverx.c
//
// C/C++ interface to DriverX kernel-mode driver services.
//
// DriverX, Version 3
//
// Copyright (C) 1995-1997, Tetradyne Software Inc.
// All rights reserved.
#include "Driverx.h"
#include "DxClass.h"
#include <stdio.h>

#ifdef _UNICODE
	#include <tchar.h>
#else
	// Support older compilers which don't have tchar.h

	#define _T(x)		x
	#define _tcscpy		strcpy	
	#define _tcsclen	strlen
	#define _stprintf	sprintf
#endif

// Some older win32 headers don't have the VER_xxx defines

#ifndef VER_PLATFORM_WIN32_WINDOWS
#define VER_PLATFORM_WIN32_WINDOWS 1
#endif

#ifndef VER_PLATFORM_WIN32_NT
#define VER_PLATFORM_WIN32_NT 2
#endif
//////////////////////////////////////////
// Global variables
static DWORD g_dwPlatform = (DWORD)(-1);
//////////////////////////////////////////
// Local Routines
static VOID GetDeviceConfig(HKEY hKeyDevice, HWDEVICE_CONFIG* pConfig);
static LONG OpenKey(HKEY hParentKey, LPCSTR pszSubkeyName, 
	REGSAM samDesired, BOOL bCreate, PHKEY phSubkey);
static LONG OpenUserDevicesKey(REGSAM samDesired, PHKEY phKeyUserDevices);
static LONG OpenDriverKey(LPCSTR pszName, REGSAM samDesired, PHKEY phKeyDriver);
static LONG SetRegistryConfiguration(HKEY hKeyDevice, 
	HWDEVICE_CONFIG* pConfig);


/////////////////// SetOsPlatform()
//
// Returns the OS VER_xxx code and stores it in g_dwPlatform.

static DWORD SetOsPlatform()
{
	OSVERSIONINFO osInfo;

	memset(&osInfo, 0, sizeof(OSVERSIONINFO));
	osInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osInfo);
	if ((osInfo.dwPlatformId != VER_PLATFORM_WIN32_NT)
		&& (osInfo.dwPlatformId != VER_PLATFORM_WIN32_WINDOWS))
	{
		TRACE(_T("DriverX: Unsupported windows platform.\n"));
		ASSERT(NULL);
	}
	else
		g_dwPlatform = osInfo.dwPlatformId;

	return g_dwPlatform;
}

/////////////////// Exported IsOnXxx() routines

BOOL _stdcall IsOnNT() 
{ 
	if (g_dwPlatform == ((DWORD)(-1)))
		SetOsPlatform();

	return (g_dwPlatform == VER_PLATFORM_WIN32_NT);
}

BOOL _stdcall IsOn95()
{
	if (g_dwPlatform == ((DWORD)(-1)))
		SetOsPlatform();

	return (g_dwPlatform == VER_PLATFORM_WIN32_WINDOWS);
}

/////////////////// HwTrace()

VOID _stdcall HwTrace(LPCSTR pszFormat, ...)
{
	va_list args;
	va_start(args, pszFormat);

	char szBuffer[512];
	vsprintf(szBuffer, pszFormat, args);
	OutputDebugString(szBuffer);

	va_end(args);
}

/////////////////// HwAssertionFailure()

VOID _stdcall HwAssertionFailure(LPCSTR pszFile, int nLine)
{
	char szBuffer[512];
	_stprintf(szBuffer, _T("DriverX Assertion Failed: File %s, Line %d\n"),
				pszFile, nLine );
	OutputDebugString(szBuffer);
	DebugBreak();
}

///////////////// HwNewDevice()
//
// Allocates and initializes a new HWDEVICE structure.
//
// Note: HwDeleteDevice must be called when done with the HWDEVICE.
//
// Parameters:
//
//		pIsr (IN) - interrupt service routine callback for device.

HWDEVICE* _stdcall HwNewDevice(VOID (*pIsr)(HWDEVICE*))
{
	HWDEVICE* pDevice = new HWDEVICE;
	if (pDevice)
		CommonConstructDevice(pDevice, pIsr);

	return pDevice;
}

/////////////////// CommonConstructDevice()
//
// Initializes a new HWDEVICE.

VOID _stdcall CommonConstructDevice(HWDEVICE* pDevice, VOID (*pIsr)(HWDEVICE*))
{
	ASSERT(pDevice);
	memset(pDevice, 0, sizeof(HWDEVICE));
	pDevice->pIsr = pIsr;
	pDevice->hDevice = INVALID_HANDLE_VALUE;
	pDevice->hDeviceIsr = INVALID_HANDLE_VALUE;

	// For synch of HwExecute and Isr
	InitializeCriticalSection(&(pDevice->crSection));
	pDevice->hCancelEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
}

/////////////////// HwSetErrorHandler()
//
// Sets the error handler for communication and ISR errors.

VOID _stdcall HwSetErrorHandler(HWDEVICE* pDevice, VOID (*pHandler)(HWDEVICE*,DWORD))
{
	ASSERT(pDevice);
	pDevice->pErrorHandler = pHandler;
}

/////////////////// HwConnectDevice()
//
// Sets up a reference from this HWDEVICE structure to an underlying kernel
// device object.
//
// Parameters:
//
//		pDevice (IN) - pointer to HWDEVICE allocated with HwNewDevice
//		pszDevice (IN) - name of kernel-mode device to connect to
//			(NULL can be passed to connect to the default DriverX device)
//
// Returns Value:
//
//		TRUE if successful, FALSE on error. 

BOOL _stdcall HwConnectDevice(HWDEVICE* pDevice, LPCSTR pszDevice)
{
	BOOL bRet;
	DWORD nDeviceID;
	DWORD cbReturned;
	TCHAR szDefault[] = _T("_Default30");

	ASSERT(pDevice);
	if (!pszDevice)
		pszDevice = szDefault;
	ASSERT(_tcsclen(pszDevice) <= MAX_NAME_LEN);
	ASSERT(pDevice->hDevice == INVALID_HANDLE_VALUE);

	if (IsOn95())
	{
		// Open handle to DriverX VxD for non ISR threads (non-overlapped I/O)
		pDevice->hDevice = CreateFile(_T("\\\\.\\DRIVERX.VXD"), 
			0, 
			0,
			NULL, 
			0, 
			0,
			NULL);

		if (pDevice->hDevice != INVALID_HANDLE_VALUE)
		{
			// Get the ID of the device from the VxD
			bRet = DeviceIoControl(pDevice->hDevice,
									IOCTL_GET_DEVICE_ID,
									(PVOID)pszDevice, 
									_tcsclen(pszDevice),
									&nDeviceID, sizeof(DWORD),
									&cbReturned,
									NULL);

			if (!bRet || nDeviceID == (DWORD)(-1))
			{
				CloseHandle(pDevice->hDevice);
				pDevice->hDevice = INVALID_HANDLE_VALUE;
			}
			else
			{
				// Open second handle for ISR thread (overlapped I/O)
				pDevice->hDeviceIsr = CreateFile(_T("\\\\.\\DRIVERX.VXD"), 
					0, 
					0,
					NULL, 
					0, 
					FILE_FLAG_OVERLAPPED,
					NULL);

				if (pDevice->hDeviceIsr == INVALID_HANDLE_VALUE)
				{
					CloseHandle(pDevice->hDevice);
					pDevice->hDevice = INVALID_HANDLE_VALUE;
				}
				else
				{
					// Success
					pDevice->nID = nDeviceID;
				}
			}
		}
	}
	else
	{
		// On NT, we have a unique handle for each kernel device.
		TCHAR szPath[MAX_NAME_LEN + 4];
		_stprintf(szPath, _T("\\\\.\\%s"), pszDevice);

		pDevice->hDevice = CreateFile(szPath, 
					GENERIC_READ|GENERIC_WRITE, 
					0, 
					NULL,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL,
					NULL); 

		if (pDevice->hDevice != INVALID_HANDLE_VALUE)
		{
			// Open second handle for ISR thread (overlapped I/O)
			pDevice->hDeviceIsr = CreateFile(szPath, 
					GENERIC_READ|GENERIC_WRITE, 
					0, 
					NULL,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
					NULL); 

			if (pDevice->hDeviceIsr == INVALID_HANDLE_VALUE)
			{
				CloseHandle(pDevice->hDevice);
				pDevice->hDevice = INVALID_HANDLE_VALUE;
			}
		}
	}

	return (pDevice->hDevice != INVALID_HANDLE_VALUE);
}

///////////////// HwDisconnectDevice()
//
// Disconnects from a kernel-mode device object (or DriverX VxD on Win95).

VOID _stdcall HwDisconnectDevice(HWDEVICE* pDevice)
{
	ASSERT(pDevice);

	if (pDevice->hDevice != INVALID_HANDLE_VALUE)
	{
		// Cancel the ISR thread (if running)
		HwDisableIsr(pDevice);

		CloseHandle(pDevice->hDevice);
		pDevice->hDevice = INVALID_HANDLE_VALUE;

		ASSERT(pDevice->hDeviceIsr != INVALID_HANDLE_VALUE);
		CloseHandle(pDevice->hDeviceIsr);
		pDevice->hDeviceIsr = INVALID_HANDLE_VALUE;
	}
}

///////////////// HwDeleteDevice()
//
// Deletes a HWDEVICE allocated using HwNewDevice().

VOID _stdcall HwDeleteDevice(HWDEVICE* pDevice)
{
	ASSERT(pDevice);
	CommonDestroyDevice(pDevice);
	delete pDevice;
}

/////////////////// CommonDestroyDevice()
//
// Cleans up a HWDEVICE.

VOID _stdcall CommonDestroyDevice(HWDEVICE *pDevice)
{
	// Close the device handle and disable ISR thread
	HwDisconnectDevice(pDevice);

	// Close the event handle for cancelling the ISR thread
	if (pDevice->hCancelEvent != NULL)
	{
		CloseHandle(pDevice->hCancelEvent);
		pDevice->hCancelEvent = NULL;
	}

	DeleteCriticalSection(&(pDevice->crSection));
}

/////////////////// HwExecute()
//
// Executes a list of one or more operations.
//
// Parameters:
//
//		pDevice (IN) - device
//		pOps (IN) - array of operations to execute in driver
//
// Return Value:
//
//		Returns TRUE if successful, FALSE on error.

BOOL _stdcall HwExecute(HWDEVICE* pDevice, PDXOPHEADER pOps)
{
	BOOL bRet;
	DWORD dwBufSize;
	DWORD cbReturned;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(pOps);

	// Just return success if there are no ops to execute
	if (!pOps->nOps)
		return TRUE;

	pOps->nDeviceID = pDevice->nID;
	dwBufSize = sizeof(DXOPHEADER) + ((pOps->nAllocated-1)*sizeof(DXOP)) + pOps->cbData;

	// Synch at top level for 3.0
	pOps->fInFlags = DXOPF_ISR_THREAD;
	EnterCriticalSection(&(pDevice->crSection));

	// Send the port operations down to the driver for execution
	bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD) IOCTL_EXECUTE_OPS,
					pOps, dwBufSize,
					pOps, dwBufSize,
					&cbReturned,
					NULL);

	LeaveCriticalSection(&(pDevice->crSection));

	if (!bRet)
	{
		TRACE(_T("DriverX: DeviceIoControl failed in HwExecute.\n"));	
		if (pDevice->pErrorHandler)
			(pDevice->pErrorHandler)(pDevice, HW_IOCTL_ERROR);
	}
	else
	{
		// Store the data for any inp operations
		PDXOP p = pOps->ops;
		DWORD i;
		for (i=0; i< pOps->nOps; i++, p++)
		{
			switch (p->fType)
			{
			case DXOP_INP:
			case DXOP_INPW:
			case DXOP_INPD:
				if (p->Inp.pVar)
					*((PDWORD)(p->Inp.pVar)) = p->Inp.nValue;
				break;
			}
		}
	}

	pOps->nOps = 0;
	return bRet;
}

/////////////////// InitializeOpHeader()
//
// Helper routine which initializes an operation queue header.

void InitializeOpHeader(DXOPHEADER* pHeader, DWORD nIndex, DWORD nAllocated, DWORD cbData)
{
	memset(pHeader, 0, sizeof(DXOPHEADER));
	pHeader->nVersion = ((DX_MAJOR_VER << 16) | (DX_MINOR_VER));
	pHeader->nIndex = nIndex;
	pHeader->nAllocated = nAllocated;
	pHeader->cbData = cbData;
	pHeader->cbDataAllocated = cbData;
}

/////////////////// HwAllocateOpQueue()
//
// Creates and initializes an operation queue.
//
// Parameters:
//
//		pIndex (IN) - target port range.
//		nOps (IN) - number of operations in this block.
//		cbData (IN) - size of data buffer for queue.

PDXOPHEADER _stdcall HwAllocateOpQueue(DWORD nIndex, DWORD nOps, DWORD cbData)
{
	int nSize;
	PDXOPHEADER pOpHeader;

	ASSERT(nOps);
	nSize = sizeof(DXOPHEADER) + (nOps-1) * sizeof(DXOP) + cbData;
	pOpHeader = (PDXOPHEADER) (new char[nSize]);
	if (!pOpHeader)
		return NULL;

	memset(pOpHeader, 0, nSize);
	InitializeOpHeader(pOpHeader, nIndex, nOps, cbData);
	return pOpHeader;
}

/////////////////// HwDeleteOpQueue()
//
// Free memory for operation queue.

VOID _stdcall HwDeleteOpQueue(PDXOPHEADER pOpHeader)
{
	if (pOpHeader)
		delete pOpHeader;
}

/////////////////// HwQueueInPort()
//
// Sets up an inp operation in an array of operations to be sent
// to the kernel-mode driver for execution.
//
// Will increment nOps for you
//
// Parameters:
//
//		pHeader (IN) - pointer to DXOPHEADER
//		nOffset (IN) - offset into port range
//		nSize (IN) - size of operation (1-byte, 2-word, 4-dword)
//		pRetData (IN) - address of input return

VOID _stdcall HwQueueInPort(PDXOPHEADER pHeader, DWORD nOffset, DWORD nSize, PULONG pRetData)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_INP_BASE + nSize;		
	pOp->Inp.nOffset = nOffset;		
	pOp->Inp.pVar = pRetData;		
	(pHeader->nOps)++;									
}

VOID _stdcall HwQueueInp(PDXOPHEADER pHeader, DWORD nOffset, PDWORD pdwReturn)
	{ HwQueueInPort(pHeader, nOffset, 1, pdwReturn); }

VOID _stdcall HwQueueInpw(PDXOPHEADER pHeader, DWORD nOffset, PDWORD pdwReturn)
	{ HwQueueInPort(pHeader, nOffset, 2, pdwReturn); }

VOID _stdcall HwQueueInpl(PDXOPHEADER pHeader, DWORD nOffset, PDWORD pdwReturn)
	{ HwQueueInPort(pHeader, nOffset, 4, pdwReturn); }

/////////////////// HwQueueOutPort()
//
// Sets up a outp operation in an array of operations to be sent
// to the kernel-mode driver for execution.
//
// Will increment nOps for you
//
// Parameters:
//
//		pHeader (IN) - pointer to DXOPHEADER
//		nOffset (IN) - offset into port range
//		nSize (IN) - type of operation (1-byte, 2-word, 4-dword)
//		nData (IN) - data for outp

VOID _stdcall HwQueueOutPort(PDXOPHEADER pHeader, DWORD nOffset, DWORD nSize, DWORD nData)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_OUTP_BASE + nSize;		
	pOp->Outp.nOffset = nOffset;		
	pOp->Outp.nValue = nData;		
	(pHeader->nOps)++;									
}

VOID _stdcall HwQueueOutp(PDXOPHEADER pHeader, DWORD nOffset, DWORD nData)
	{ HwQueueOutPort(pHeader, nOffset, 1, nData); }

VOID _stdcall HwQueueOutpw(PDXOPHEADER pHeader, DWORD nOffset, DWORD nData)
	{ HwQueueOutPort(pHeader, nOffset, 2, nData); }

VOID _stdcall HwQueueOutpl(PDXOPHEADER pHeader, DWORD nOffset, DWORD nData)
	{ HwQueueOutPort(pHeader, nOffset, 4, nData); }

/////////////////// HwQueueBufInPort()
//
// Sets up a inp operation in an array of operations to be sent
// to the kernel-mode driver for execution.

VOID _stdcall HwQueueBufInPort(PDXOPHEADER pHeader, DWORD nOffset, DWORD nSize)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_BUF_INP_BASE + nSize;		
	pOp->Inp.nOffset = nOffset;		
	(pHeader->nOps)++;									
}

VOID _stdcall HwQueueBufInp(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueBufInPort(pHeader, nOffset, 1); }

VOID _stdcall HwQueueBufInpw(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueBufInPort(pHeader, nOffset, 2); }

VOID _stdcall HwQueueBufInpl(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueBufInPort(pHeader, nOffset, 4); }

/////////////////// HwQueueBufOutp()
//
// Sets up a outp operation in an array of operations to be sent
// to the kernel-mode driver for execution.

VOID _stdcall HwQueueBufOutPort(PDXOPHEADER pHeader, DWORD nOffset, DWORD nSize)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_BUF_OUTP_BASE + nSize;
	pOp->Outp.nOffset = nOffset;		
	(pHeader->nOps)++;									
}

VOID _stdcall HwQueueBufOutp(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueBufOutPort(pHeader, nOffset, 1); }

VOID _stdcall HwQueueBufOutpw(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueBufOutPort(pHeader, nOffset, 2); }

VOID _stdcall HwQueueBufOutpl(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueBufOutPort(pHeader, nOffset, 4); }

/////////////////// HwQueueTempOutp()
//
// Sets up a queued outp of the temporary variable value.

VOID _stdcall HwQueueTempOutPort(PDXOPHEADER pHeader, DWORD nOffset, DWORD nSize)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_TEMP_OUTP_BASE + nSize;
	pOp->Outp.nOffset = nOffset;		
	(pHeader->nOps)++;									
}

VOID _stdcall HwQueueTempOutp(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueTempOutPort(pHeader, nOffset, 1); }

VOID _stdcall HwQueueTempOutpw(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueTempOutPort(pHeader, nOffset, 2); }

VOID _stdcall HwQueueTempOutpl(PDXOPHEADER pHeader, DWORD nOffset)
	{ HwQueueTempOutPort(pHeader, nOffset, 4); }

/////////////////// HwQueueAndTemp()


VOID _stdcall HwQueueAndTemp(PDXOPHEADER pHeader, DWORD nMask)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_AND_TEMP;
	pOp->Value.nValue1 = nMask;		
	(pHeader->nOps)++;									
}

/////////////////// HwQueueOrTemp()

VOID _stdcall HwQueueOrTemp(PDXOPHEADER pHeader, DWORD nMask)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_OR_TEMP;
	pOp->Value.nValue1 = nMask;		
	(pHeader->nOps)++;									
}

/////////////////// HwQueueDelay()

VOID _stdcall HwQueueDelay(PDXOPHEADER pHeader, DWORD nMicroSec)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_DELAY;
	pOp->Value.nValue1 = nMicroSec;		
	(pHeader->nOps)++;									
}

/////////////////// HwGetQueueBuffer()
//
// Returns a pointer to the data part of OpHeader 

PVOID _stdcall HwGetQueueBuffer(PDXOPHEADER pHeader)
{
	if (pHeader)
		return ((PUCHAR)pHeader) + sizeof(DXOPHEADER) 
				+ (pHeader->nAllocated -1) * sizeof(DXOP);
	else
		return NULL;
}

/////////////////// HwGetQueueResults()
//
// Returns the results after executing a queue of operations.

ULONG _stdcall HwGetQueueResults(PDXOPHEADER pHeader, PDWORD pcbProcessed)
{
	ASSERT(pHeader);
	if (pcbProcessed)
		*pcbProcessed = pHeader->cbReturn;

	return pHeader->fOutFlags;
}

/////////////////// HwQueueWhileInpAnd()
//
// Fill header for while loop.

VOID _stdcall HwQueueWhileInpAnd(PDXOPHEADER pHeader, DWORD nOffset, 
				DWORD nDataSize, DWORD nMask, BOOL bLoopWhile, DWORD nLimit)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_WHILE_INP_AND;		
	pOp->WIA.nOffset = nOffset;		
	pOp->WIA.nTimeout = nLimit;		
	pOp->WIA.nMask = nMask;		
	pOp->WIA.nDataSize = nDataSize;		
	pOp->WIA.bLoopWhile = bLoopWhile;		
	(pHeader->nOps)++;
}

/////////////////// HwQueueWhileInpEqual()

VOID _stdcall HwQueueWhileInpEqual(PDXOPHEADER pHeader, DWORD nOffset, 
	DWORD nDataSize, DWORD nMask, DWORD nValue, BOOL bLoopWhile, DWORD nLimit)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_WHILE_INP_EQU;		
	pOp->WIE.nOffset = nOffset;		
	pOp->WIE.nTimeout = nLimit;		
	pOp->WIE.nMask = nMask;		
	pOp->WIE.nValue = nValue;
	pOp->WIE.nDataSize = nDataSize;		
	pOp->WIE.bLoopWhile = bLoopWhile;		
	(pHeader->nOps)++;
}

////////////////// HwQueueLoopBack();
//
// Fill header for buffered data loop 

VOID _stdcall HwQueueLoopBack(PDXOPHEADER pHeader)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_LOOP_BACK;		
	(pHeader->nOps)++;
}

////////////////// HwQueueBreakIfAnd();
//
// Fill header for break out of loop 

VOID _stdcall HwQueueBreakIfAnd(PDXOPHEADER pHeader, DWORD nMask, BOOL bBreak)
{
	ASSERT(pHeader);
	ASSERT(pHeader->nOps < pHeader->nAllocated);

	PDXOP pOp = &(pHeader->ops[pHeader->nOps]);
	pOp->fType = DXOP_BREAK_IF_AND;		
	pOp->BIA.nMask = nMask;		
	pOp->BIA.bBreak = bBreak;		
	(pHeader->nOps)++;
}

/////////////////// HwEnableQueueOption()
//
// Enables a operation queue option flag.

BOOL _stdcall HwEnableQueueOption(PDXOPHEADER pHeader, DWORD fOption, BOOL bEnable)
{
	ASSERT(pHeader);

	// Currently, only one flag is valid 
	if (fOption != DXOPF_DISABLE_INTS)
		return FALSE;

	if (bEnable)
		pHeader->fInFlags |= fOption;
	else
		pHeader->fInFlags &= ~(fOption);

	return TRUE;
}

/////////////////// HwEnableIsr()
//
// Enables user-defined interrupt service routine callback. A thread 
// should be created by the caller in a suspended state. This routine 
// will resume the thread after some setup. 
//
// Parameters:
//
//		pDevice - (IN) device whose ISR is to be enabled
//		hThread - (IN) handle to thread created by user for ISR callbacks.
//		nThreadID - (IN) id of thread
//
// Return Value:
//
//		TRUE if interrupt is properly setup on driver side, else FALSE
//		if interrupt is not setup or other error occurs. If this routine
//		returns FALSE, user should close handle to thread.

BOOL _stdcall HwEnableIsr(HWDEVICE* pDevice, HANDLE hThread, DWORD nThreadID)
{
	DWORD nIntConnected;
	BOOL bRet;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(hThread);
	ASSERT(pDevice->hIsrThread == NULL);
	ASSERT(pDevice->pIsr);

	// See if the device has hooked an interrupt on the driver side
	bRet = HwGetProperty(pDevice, DXPROP_IS_INTR_CONNECTED, &nIntConnected);
	if (bRet)
	{
		if (!nIntConnected)
		{
			TRACE(_T("DriverX: No interrupt hooked on driver side.\n"));
			bRet = FALSE;
		}
		else
			bRet = HwSetProperty(pDevice, DXPROP_ISR_ENABLED, TRUE);
	}
	
	if (!bRet)
	{
		// Error. Resume thread, but set cancelEvent so thread exits.
		HwSetProperty(pDevice, DXPROP_ISR_ENABLED, FALSE);
		SetEvent(pDevice->hCancelEvent);
		ResumeThread(hThread);

		// Wait for ISR thread to exit
		if (WaitForSingleObject(hThread, 4000) != WAIT_OBJECT_0)
			TRACE(_T("DriverX: Timeout on wait for ISR thread exit.\n"));
	}
	else
	{
		// Store thread info before letting thread run
		pDevice->hIsrThread = hThread;
		pDevice->nIsrThreadID = nThreadID;
		
		// Let ISR thread run
		ResetEvent(pDevice->hCancelEvent);
		if (ResumeThread(hThread) == (DWORD)(-1))
		{
			TRACE(_T("DriverX: Could not resume ISR thread.\n"));

			HwSetProperty(pDevice, DXPROP_ISR_ENABLED, FALSE);
			pDevice->hIsrThread = NULL;
			pDevice->nIsrThreadID = 0;
			bRet = FALSE;
		}
	}

	return bRet;
}

/////////////////// HwDisableIsr()
//
// Sets the cancel event to terminate the ISR thread

VOID _stdcall HwDisableIsr(HWDEVICE* pDevice)
{
	ASSERT(pDevice);

	if (pDevice->hIsrThread)
	{
		DWORD dwWait;

		// Signal cancel event
		ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
		ASSERT(pDevice->hCancelEvent);
		SetEvent(pDevice->hCancelEvent);

		// Wait for thread to exit
		dwWait = WaitForSingleObject(pDevice->hIsrThread, 4000);
		if (dwWait != WAIT_OBJECT_0)
			TRACE(_T("DriverX: Timed out waiting for ISR thread to exit.\n"));

		HwSetProperty(pDevice, DXPROP_ISR_ENABLED, FALSE);
		pDevice->hIsrThread = NULL;
		pDevice->nIsrThreadID = 0;
	}
}

/////////////////// HwIsrThreadProc() 
//
// ISR thread procedure which continually waits on interrupts in the
// kernel-mode driver and calls the user's ISR callback. This routine
// returns on a driver error, or the cancel event being set.
//
// Return Value:
//
//		Returns 0 if cancel event is signaled, else a >0 error code.

UINT _stdcall HwIsrThreadProc(PVOID pContext)
{
	OVERLAPPED ol;
	HANDLE objectsArray[2];
	DWORD cbReturned;
	DWORD dwReturn = 0;
	BOOL bRet;
	BOOL bDone = FALSE;
	HWDEVICE* pDevice = (HWDEVICE*) pContext;

	ASSERT(pDevice);
	ASSERT(pDevice->pIsr); 
	ASSERT(pDevice->hCancelEvent);

	// Create event object for overlapped IO synchronization
	memset(&ol, 0, sizeof(OVERLAPPED));
	ol.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (ol.hEvent == NULL)
	{
		TRACE(_T("DriverX: Unable to create event in thread proc.\n"));
		if (pDevice->pErrorHandler)
			(pDevice->pErrorHandler)(pDevice, HW_EVENT_ERROR);
		return HW_EVENT_ERROR;
	}

	// Block on overlapped structure event and cancel event
	objectsArray[0] = pDevice->hCancelEvent;
	objectsArray[1] = ol.hEvent;

	// See if ISR thread has already been shut down
	if (WaitForSingleObject(pDevice->hCancelEvent, 0) == WAIT_OBJECT_0)
	{
		CloseHandle(ol.hEvent);
		return 0;
	}

	// Loop waiting for interrupts and calling ISR callback
	while (!bDone)
	{
		ResetEvent(ol.hEvent);

		bRet = DeviceIoControl(
					pDevice->hDeviceIsr,
					(DWORD) IOCTL_WAIT_INTERRUPT,
					&(pDevice->nID), sizeof(DWORD),
					&(pDevice->waitOutput), sizeof(WAIT_INTERRUPT_OUTPUT),
					&cbReturned, &ol);

		if (bRet)
		{
			// Call completed OK. Check cancel event before continuing
			if (WaitForSingleObject(pDevice->hCancelEvent, 0) == WAIT_OBJECT_0)
				bDone = TRUE;
		}
		else
		{
			// Check for proper overlapped return
			if (GetLastError() != ERROR_IO_PENDING)
			{
				TRACE(_T("DriverX: IOCTL_WAIT_INTERRUPT error.\n"));
				dwReturn = HW_IOCTL_ERROR;
				bDone = TRUE;
			}
			else
			{
				// Wait for interrupt to occur, or cancel event to be set.
				dwReturn = WaitForMultipleObjects(2, objectsArray,
					FALSE, INFINITE );

				if ((dwReturn-WAIT_OBJECT_0) == 1)
				{
					// IOCTL_WAIT_INTERRUPT finished
					GetOverlappedResult(pDevice->hDeviceIsr, &ol, 
						&cbReturned, FALSE);
				}
				else if ((dwReturn-WAIT_OBJECT_0) == 0)
				{
					// Cancel event set. End the thread.
					bDone = TRUE;
				}
				else
				{  
					DWORD dwError = GetLastError();
					TRACE(_T("DriverX: ISR WaitObjects error: %d, 0x%x\n"),
						dwError, dwError);
					dwReturn = HW_WAIT_INTR_ERROR;
					bDone = TRUE;
				}
			}
		}

		if (!bDone)
		{
			if (cbReturned != sizeof(WAIT_INTERRUPT_OUTPUT))
			{
				TRACE(_T("DriverX: IOCTL_WAIT_INTERRUPT cbReturned error.\n"));
				dwReturn = HW_WAIT_INTR_ERROR;
				bDone = TRUE;
			}
			else
			{
				// Enter device critical section and call ISR callback
				EnterCriticalSection(&(pDevice->crSection));

				try
				{
					(pDevice->pIsr)(pDevice);
				}
				catch(...)
				{
					// Leave critical section
					LeaveCriticalSection(&(pDevice->crSection));

					// Cleanup
					CloseHandle(ol.hEvent);
					HwSetProperty(pDevice, DXPROP_ISR_ENABLED, FALSE);

					// Pass on exception
					throw;
				}

				LeaveCriticalSection(&(pDevice->crSection));
			}
		}
	}

	// Cleanup
	CloseHandle(ol.hEvent);
	HwSetProperty(pDevice, DXPROP_ISR_ENABLED, FALSE);

	// Call error handler on an error
	if (dwReturn && pDevice->pErrorHandler)
		(pDevice->pErrorHandler)(pDevice, HW_EVENT_ERROR);

	return dwReturn; 
}

/////////////////// Thread proc wrappers for MFC and BC++

UINT HwMfcIsrThreadProc(LPVOID pParam)
	{ return HwIsrThreadProc(pParam); }

VOID HwBcppIsrThreadProc(LPVOID pParam)
	{ HwIsrThreadProc(pParam); }

/////////////////// HwSetProperty()
//
// Sets a device property.
//
// Return Value:
//
//        Returns True if devices supports specified property; else False.

BOOL _stdcall HwSetProperty(HWDEVICE* pDevice, DWORD nCode, DWORD nState)
{
	DWORD cbReturned;
	PROPERTY_INPUT input;
	BOOL bRet;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);

	memset(&input, 0, sizeof(PROPERTY_INPUT));
	input.nDeviceID = pDevice->nID;
	input.nPropCode = nCode;
	input.nPropState = nState;

	bRet = DeviceIoControl(pDevice->hDevice,
				IOCTL_SET_PROPERTY,
				&input, sizeof(PROPERTY_INPUT),
				NULL, 0,
				&cbReturned,
				NULL);

	if (!bRet)
		TRACE(_T("DriverX: IOCTL_SET_PROPERTY failed.\n"));

	return bRet;
}

/////////////////// HwGetProperty()
//
// Gets a device property.
//  
//     nCode (IN) -property to get.
//     pnState (OUT) -value of property returned here.
//
// Return Value:
//
//        Returns True if devices supports specified property; else False.

BOOL _stdcall HwGetProperty(HWDEVICE* pDevice, DWORD nCode, PDWORD pnState)
{
	DWORD cbReturned;
	PROPERTY_INPUT input;
	BOOL bRet;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);

	memset(&input, 0, sizeof(PROPERTY_INPUT));
	input.nDeviceID	= pDevice->nID;
	input.nPropCode	= nCode;

	bRet = DeviceIoControl(pDevice->hDevice,
				IOCTL_GET_PROPERTY,
				&input, sizeof(PROPERTY_INPUT),
				pnState, sizeof(DWORD),
				&cbReturned,
				NULL);

	if (!bRet)
		TRACE(_T("DriverX: IOCTL_GET_PROPERTY failed.\n"));

	return bRet;
}

/////////////////////////////////////////////////////////////////////////////
// HwNewMemory()
//
// Allocates and initializes a new HWMEMORY structure.

HWMEMORY* _stdcall HwNewMemory()
{
	HWMEMORY* pMem;

	pMem = new HWMEMORY;
	if (pMem)
		CommonConstructMemory(pMem);

	return pMem;
}

VOID CommonConstructMemory(HWMEMORY* pMem)
{
	memset(pMem, 0, sizeof(HWMEMORY));
}

/////////////////// HwDeleteMemory()

VOID _stdcall HwDeleteMemory(HWMEMORY* pMem)
{
	ASSERT(pMem);
	HwUnmapMemory(pMem);
	delete pMem;
}

/////////////////// HwMapDeviceMemory()
//
// HwMapDeviceMemory maps a pre-defined device memory range defined
// by a device's registry configuration/INF file/or PCI configuration. 
// If the registry or INF file MemIndex value is defined, then the 
// memory range specified by the MemIndex parameter is mapped.

BOOL _stdcall HwMapDeviceMemory(HWMEMORY* pMem, HWDEVICE* pDevice)
{
	ASSERT(pMem);
	ASSERT(pDevice);
	HwUnmapMemory(pMem);

	pMem->pDevice = pDevice;

	// If type has not already been set by calling HwSetMemIndex...
	if (LOWORD(pMem->fType) != DXMEM_TYPE_DEVICE2)
		pMem->fType = DXMEM_TYPE_DEVICE;

	return HwMapMemoryEx(pMem, 0, 0, 0, 0);
}

/////////////////// HwMapDmaBuffer()

BOOL _stdcall HwMapDmaBuffer(HWMEMORY* pMem, HWDEVICE* pDevice)
{
	ASSERT(pMem);
	ASSERT(pDevice);
	HwUnmapMemory(pMem);

	pMem->pDevice = pDevice;
	pMem->fType = DXMEM_TYPE_DMA_BUFFER;
	return HwMapMemoryEx(pMem, 0, 0, 0, 0);
}

/////////////////// HwMapMemory()
//
// Maps a range of physical addresses into the callers processes virtual
// address space.
//
// Parameters:
//
//		pMem (IN) - HWMEMORY structure allocated with HwNewMemory.
//		nBaseAddress (IN) - base physical address
//		nCount (IN) - length of physical address range to map.
//
// Return Value:
//
//		Returns TRUE if successful, FALSE on error.

BOOL _stdcall HwMapMemory(HWMEMORY* pMem, DWORD nBaseAddress, DWORD nCount)
{
	ASSERT(pMem);
	HwUnmapMemory(pMem);

	pMem->pDevice = NULL;
	pMem->fType = DXMEM_TYPE_PHYSICAL;
	return HwMapMemoryEx(pMem, nBaseAddress, nCount, 1 /* Isa */, 0);
}

/////////////////// HwMapMemoryEx()
//
// Helper routine.

BOOL _stdcall HwMapMemoryEx(HWMEMORY* pMem, DWORD nBaseAddress, DWORD nCount,
		DWORD fInterfaceType, DWORD nBusNumber)
{
	ASSERT(pMem);
	HwUnmapMemory(pMem);

	HWDEVICE* pDevice;
	if (pMem->pDevice)
		pDevice = pMem->pDevice;
	else
	{
		// Go through _Default30 device
		pDevice = HwNewDevice(NULL);
		if (!HwConnectDevice(pDevice, NULL))
		{
			HwDeleteDevice(pDevice);
			return FALSE;
		}
	}

	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);

	MAPMEM_INPUT input;
	memset(&input, 0, sizeof(MAPMEM_INPUT));

	input.nDeviceID			= pDevice->nID;
	input.fType				= LOWORD(pMem->fType);
	input.fInterfaceType	= fInterfaceType;
	input.nBusNumber		= nBusNumber;
	input.nAddress			= nBaseAddress;
	input.nLength			= nCount;
	input.fAddressSpace		= 0;	// memory space
	input.nIndex			= HIWORD(pMem->fType);

	memset(&pMem->output, 0, sizeof(MAPMEM_OUTPUT));

	DWORD cbReturned;
	BOOL bRet;
	bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_MAP_MEMORY,
					&input,
					sizeof(MAPMEM_INPUT),
					&pMem->output,
					sizeof(MAPMEM_OUTPUT),
					&cbReturned,
					NULL);

	if (!bRet || (cbReturned != sizeof(MAPMEM_OUTPUT)))
	{
		TRACE(_T("DriverX: IOCTL_MAP_MEMORY DeviceIoControl failed!\n"));		
		bRet = FALSE;
	}

	if (pMem->output.nVirtualAddress == 0)
	{
		TRACE(_T("DriverX: Mapping of address failed.\n"));
		bRet = FALSE;
	}

	// Close handle to _Default30 device if used
	if (pDevice != pMem->pDevice)
		HwDeleteDevice(pDevice);

	return bRet;
}

/////////////////// HwGetMemoryPointer()
//
// Returns virtual address accessible by the app.

PBYTE _stdcall HwGetMemoryPointer(HWMEMORY* pMem)
{
	ASSERT(pMem);
	if (!pMem->output.nVirtualAddress)
		TRACE(_T("DriverX: Warning - returning NULL from HwGetMemoryPointer.\n"));
	return (PBYTE) pMem->output.nVirtualAddress;
}

/////////////////// HwGetPhysicalAddress()
//
// Returns physical address of mapped memory range (usually used for
// DMA buffers)

DWORD _stdcall HwGetPhysicalAddress(HWMEMORY* pMem)
{
	ASSERT(pMem);
	if (!pMem->output.nPhysicalAddress)
		TRACE(_T("DriverX: Warning - returning 0 from HwGetPhysicalAddress.\n"));
	return pMem->output.nPhysicalAddress;
}

/////////////////// HwUnmapMemory()
//
// Unmaps memory previously mapped using HwMapXxxMemory.

BOOL _stdcall HwUnmapMemory(HWMEMORY* pMem)
{
	ASSERT(pMem);

	// Just return if no memory is mapped through this HWMEMORY
	if (!pMem->output.nVirtualAddress)
		return TRUE; 

	HWDEVICE* pDevice;
	if (pMem->pDevice)
		pDevice = pMem->pDevice;
	else
	{
		// Go through _Default30 device
		pDevice = HwNewDevice(NULL);
		if (!HwConnectDevice(pDevice, NULL))
		{
			HwDeleteDevice(pDevice);
			return FALSE;
		}
	}

	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);

	UNMAPMEM_INPUT input;
	memset(&input, 0, sizeof(UNMAPMEM_INPUT));

	input.nDeviceID = pDevice->nID;
	input.fType		= LOWORD(pMem->fType);
	input.nIndex	= HIWORD(pMem->fType);

	ASSERT(pMem->output.nDriverVirtualAddress);
	input.nDriverVirtualAddress = pMem->output.nDriverVirtualAddress;

	DWORD cbReturned;
	BOOL bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_UNMAP_MEMORY,
					&input,
					sizeof(UNMAPMEM_INPUT),
					NULL,
					0,
					&cbReturned,
					NULL);

	if (!bRet)
		TRACE(_T("DriverX: UNMAP_MEMORY DeviceIoControl failed!"));		

	memset(&pMem->output, 0, sizeof(MAPMEM_OUTPUT));
	if (pDevice != pMem->pDevice)
		HwDeleteDevice(pDevice);

	return bRet;
}

/////////////////// HwReadMemory()
//
// This routine is for VB apps to read physical memory mapped using
// HwMapMemory. C users can just directly use the pointer returned
// from HwGetMemoryPointer.

VOID _stdcall HwReadMemory(HWMEMORY* pMem, DWORD nOffset, 
		LPBYTE pData, DWORD nCount)
{
	ASSERT(pData);
	ASSERT((nOffset + nCount) <= pMem->output.nCount);
	memcpy(pData, HwGetMemoryPointer(pMem) + nOffset, nCount);
}

BYTE _stdcall HwReadByteMem(HWMEMORY* pMem, DWORD nOffset)
{
	BYTE b;
	HwReadMemory(pMem, nOffset, &b, 1);
	return b;
}

WORD _stdcall HwReadWordMem(HWMEMORY* pMem, DWORD nOffset)
{
	WORD w;
	HwReadMemory(pMem, nOffset * sizeof(WORD), (PBYTE) &w, sizeof(WORD));
	return w;
}

DWORD _stdcall HwReadLongMem(HWMEMORY* pMem, DWORD nOffset)
{
	DWORD dw;
	HwReadMemory(pMem, nOffset * sizeof(DWORD), (PBYTE) &dw, sizeof(DWORD));
	return dw;
}

/////////////////// HwWriteMemory()
//
// This routine is for VB apps to read physical memory mapped using
// HwMapMemory. C users can just directly use the pointer returned
// from HwGetMemoryPointer.

VOID _stdcall HwWriteMemory(HWMEMORY* pMem, DWORD nOffset, 
		LPBYTE pData, DWORD nCount)
{
	ASSERT(pData);
	ASSERT((nOffset + nCount) <= pMem->output.nCount);
	memcpy(HwGetMemoryPointer(pMem) + nOffset, pData, nCount);
}

VOID _stdcall HwWriteByteMem(HWMEMORY* pMem, DWORD nOffset, BYTE b)
	{ HwWriteMemory(pMem, nOffset, &b, 1); }

VOID _stdcall HwWriteWordMem(HWMEMORY* pMem, DWORD nOffset, WORD w)
	{ HwWriteMemory(pMem, nOffset * sizeof(WORD), (PBYTE) &w, sizeof(WORD)); }

VOID _stdcall HwWriteLongMem(HWMEMORY* pMem, DWORD nOffset, DWORD dw)
	{ HwWriteMemory(pMem, nOffset * sizeof(DWORD), (PBYTE) &dw, sizeof(DWORD)); }

/////////////////// HwGetMemoryIndex()

DWORD _stdcall HwGetMemoryIndex(HWMEMORY* pMem)
{ 
	ASSERT(pMem);
	return HIWORD(pMem->fType); 
}

/////////////////// HwSetMemoryIndex()

VOID _stdcall HwSetMemoryIndex(HWMEMORY* pMem, DWORD dwIndex)
{ 
	ASSERT(pMem);
	ASSERT(pMem->output.nVirtualAddress == 0); // Need to unmap before changing

	// Set type to DXMEM_TYPE_DEVICE2 so driver uses nIndex parameter
	pMem->fType = DXMEM_TYPE_DEVICE2 | (dwIndex << 16); 
}

/////////////////// HwConnectInterrupt()
//
// This call hooks the kernel device associated with the given HWDEVICE to
// an interrupt. If this call is successful, then a call to HwEnableIsr
// can then be made to enable the user-mode interrupt service routine
// for the HWDEVICE (passed to HwNewDevice).
//
// Note: before calling this routine, a successful call to HwConnectDevice
// must have been made to associate a kernel-mode device with this HWDEVICE.
//
// Return Value:
//
//		Returns TRUE if successful, FALSE on error.

BOOL _stdcall HwConnectInterrupt(HWDEVICE* pDevice, DWORD nInterruptLevel)
{
	return HwConnectInterruptEx( pDevice, nInterruptLevel, nInterruptLevel,
				1 /* Isa */, 0 );
}

BOOL _stdcall HwConnectInterruptEx(HWDEVICE* pDevice, DWORD nInterruptLevel,
	DWORD nVector, DWORD fInterfaceType, DWORD nBusNumber)
{
	CONNECT_INTERRUPT_INPUT input;
	CONNECT_INTERRUPT_OUTPUT output;
	DWORD cbReturned;
	BOOL bRet;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(pDevice->hIsrThread == NULL); // need to disable ISR first

	if (nVector == (DWORD)(-1))
		nVector = nInterruptLevel;

	input.nDeviceID 		= pDevice->nID;
	input.fInterfaceType 	= fInterfaceType;
	input.nBusNumber 		= nBusNumber;
	input.nInterruptLevel 	= nInterruptLevel;
	input.nInterruptVector 	= nVector;
	input.fInterruptMode 	= 1; // latched

	memset(&output, 0, sizeof(CONNECT_INTERRUPT_OUTPUT));

	bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_CONNECT_INTERRUPT,
					&input,
					sizeof(CONNECT_INTERRUPT_INPUT),
					&output,
					sizeof(CONNECT_INTERRUPT_OUTPUT),
					&cbReturned,
					NULL);

	if (cbReturned != sizeof(CONNECT_INTERRUPT_OUTPUT))
		bRet = FALSE;

	return bRet;
}

/////////////////// HwDisconnectInterrupt()
//
// Unhooks the kernel device associated with a given HWDEVICE from
// any interrupt it may be connected with.

BOOL _stdcall HwDisconnectInterrupt(HWDEVICE* pDevice)
{
	DWORD cbReturned;
	DWORD dwID;
	BOOL bRet;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(pDevice->hIsrThread == NULL); // Need to disable ISR first

	dwID = pDevice->nID;
	bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_DISCONNECT_INTERRUPT,
					&dwID,
					sizeof(DWORD),
					NULL,
					0,
					&cbReturned,
					NULL );

	return bRet;
}

/////////////////// HwGetIsrData()
//
// This routine should only be used if a device is configured for an ISR
// port operation (see the IsrOp value help).
//
// Return Value:
//
//		Port data for any ISR op port read operations.

DWORD _stdcall HwGetIsrData(HWDEVICE* pDevice)
{
	ASSERT(pDevice);
	return pDevice->waitOutput.nData;
}

/////////////////// HwGetIsrIdData()
//
// Returns data from any ISR ID operation defined by a device 
// configuration.

DWORD _stdcall HwGetIsrIdData(HWDEVICE* pDevice)
{
	ASSERT(pDevice);
	return pDevice->waitOutput.nIdData;
}

/////////////////// HwBeginIsrSynch()

VOID _stdcall HwBeginIsrSynch(HWDEVICE* pDevice)
{
	ASSERT(pDevice);
	EnterCriticalSection(&pDevice->crSection);
}

/////////////////// HwEndIsrSynch()

VOID _stdcall HwEndIsrSynch(HWDEVICE* pDevice)
{
	ASSERT(pDevice);
	LeaveCriticalSection(&pDevice->crSection);
}

/////////////////// HwConfigureDevice()
//
// For pre-3.3 compatibility. See HwConfigureDeviceEx.

BOOL _stdcall HwConfigureDevice(LPCSTR pszName, HWDEVICE_CONFIG* pConfig)
{
    ASSERT(pConfig);

	// pConfig may point to an old pre-3.3 structure. Copy into new structure.
	HWDEVICE_CONFIG config;
	memset(&config, 0, sizeof(HWDEVICE_CONFIG));
	config.cbSize = sizeof(HWDEVICE_CONFIG);

	// Copy configuration
	memcpy(&config, pConfig, HWDEVICE_CONFIG_PRE33_SIZE);

	// Call new ...Ex routine
	return HwConfigureDeviceEx(pszName, &config);
}

/////////////////// HwConfigureDeviceEx()
//
// Sets a device's configuration registry value entries.
//
// On NT, user device keys are held under
//
//		HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services
//			\Drivers\DriverX\Parameters
//
// On Windows 95 user device keys are under
//
//
//		HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services
//			\VxD\DRIVERX
//
// Return Value:
//
//		Returns TRUE if successful, FALSE on a registry access or
//      invalid configuration structure error.

BOOL _stdcall HwConfigureDeviceEx(LPCSTR pszName, HWDEVICE_CONFIG* pConfig)
{
	HKEY hKeyUserDevices;
	HKEY hKeyDevice;
	LONG fRet;

	ASSERT(pConfig);
    ASSERT(pszName);

	// Open the key under which user device configurations are stored
	if (OpenUserDevicesKey(KEY_ALL_ACCESS, &hKeyUserDevices) != ERROR_SUCCESS)
		return FALSE;

	// Create or open the subkey with the given device name
	fRet = OpenKey(hKeyUserDevices, pszName, KEY_ALL_ACCESS, 
				TRUE, &hKeyDevice);
	RegCloseKey(hKeyUserDevices);
	if (fRet != ERROR_SUCCESS)
		return FALSE;

	// Set the configuration values under the device key
	SetRegistryConfiguration(hKeyDevice, pConfig);

	RegFlushKey(hKeyDevice);
	RegCloseKey(hKeyDevice);

	return TRUE;
}

/////////////////// HwConfigureDriver()
//
// Configures the supporting device driver in the system registry. 
//
// Returns TRUE if successful, FALSE on error.

BOOL _stdcall HwConfigureDriver()
{
	HKEY 	hKeyDrvr, hKeyParameters;
	LONG 	fRet;
	DWORD 	dwValue;
	TCHAR	buf[256];
	
	if (IsOn95())
	{
		// Create key for VxD
		fRet = OpenKey(HKEY_LOCAL_MACHINE, 
				_T("System\\CurrentControlSet\\Services\\VxD\\DRIVERX"),
				KEY_ALL_ACCESS,
				TRUE,
				&hKeyDrvr);

		if (fRet != ERROR_SUCCESS)
			return FALSE;

		// Configure VxD for static load
		_tcscpy(buf, _T("DRIVERX.VXD"));
		fRet = RegSetValueEx(hKeyDrvr,
						_T("StaticVxD"),
						0,
						REG_SZ,
						(CONST BYTE*)buf,
						_tcsclen(buf));

		if (fRet != ERROR_SUCCESS)
		{
			RegCloseKey(hKeyDrvr);
			return FALSE;
		}

		RegFlushKey(hKeyDrvr);
		RegCloseKey(hKeyDrvr);
	}
	else
	{
		SC_HANDLE hSCManager = OpenSCManager(
			NULL, 
			NULL,
			SC_MANAGER_ALL_ACCESS);

		if (hSCManager == NULL)
		{
			// User probably does not have admin rights
			TRACE(_T("DriverX: OpenSCManager failed.\n"));
			return FALSE;
		}

		// Create DriverX service key
		SC_HANDLE hService = CreateService(
			hSCManager,
			_T("DriverX"),
			_T("DriverX"),
			SERVICE_ALL_ACCESS,
			SERVICE_KERNEL_DRIVER,
			SERVICE_AUTO_START,
			SERVICE_ERROR_NORMAL,
			_T("\\SystemRoot\\System32\\Drivers\\driverx.sys"),
			NULL,
			NULL,
			NULL,
			NULL,
			NULL);

		CloseServiceHandle(hSCManager);
		if (hService == NULL)
		{
			DWORD dwErr = GetLastError();
			if (dwErr != ERROR_SERVICE_EXISTS)
			{
				TRACE(_T("DriverX: CreateService failed. GetLastError=%d(0x%x)\n"), dwErr, dwErr);
				return FALSE;
			}
		}
		else
			CloseServiceHandle(hService);

		fRet = OpenKey(HKEY_LOCAL_MACHINE, 
				_T("System\\CurrentControlSet\\Services\\DriverX"),
				KEY_ALL_ACCESS,
				TRUE,
				&hKeyDrvr);

		if (fRet != ERROR_SUCCESS)
			return FALSE;

		// Create Parameters key under driver key
		fRet = OpenKey(hKeyDrvr, _T("Parameters"), KEY_ALL_ACCESS, 
			TRUE, &hKeyParameters);
		RegCloseKey(hKeyDrvr);
	 	if (fRet != ERROR_SUCCESS)
			return FALSE;

		// Add event logging entries
		fRet = OpenKey(HKEY_LOCAL_MACHINE, 
			_T("System\\CurrentControlSet\\Services\\EventLog\\System\\DriverX"),
			KEY_ALL_ACCESS,
			TRUE,
			&hKeyDrvr);

		if (fRet != ERROR_SUCCESS)
			return fRet;

		// Add standard driver key values
		_tcscpy(buf, _T("%SystemRoot%\\System32\\IoLogMsg.dll;%SystemRoot%\\System32\\Drivers\\DriverX.sys"));
		fRet = RegSetValueEx(hKeyDrvr,
						_T("EventMessageFile"),
						0,
						REG_EXPAND_SZ,
						(unsigned char*)buf,
						_tcsclen(buf));

		if (fRet != ERROR_SUCCESS)
		{
			RegCloseKey(hKeyDrvr);
			return FALSE;
		}

		dwValue = 0x07;	
		fRet = RegSetValueEx(hKeyDrvr,
						_T("TypesSupported"),
						0,
						REG_DWORD,
						(const unsigned char*) &dwValue,
						sizeof(DWORD));

		if (fRet != ERROR_SUCCESS)
		{
			RegCloseKey(hKeyDrvr);
			return FALSE;
		}

		RegFlushKey(hKeyDrvr);
		RegCloseKey(hKeyDrvr);
	}

	return TRUE;
}

/////////////////// HwDeleteDeviceConfiguration()
//
// Deletes a user registry device configuration of a given name.
//
// Return Value:
//
//		TRUE if successful, FALSE if unable to delete the device key.

BOOL _stdcall HwDeleteDeviceConfiguration(LPCSTR pszName)
{
	BOOL bRet;
	HKEY hKeyDevices;
	LONG fKeyRet;

	ASSERT(pszName);

	// Open driver key
	fKeyRet = OpenUserDevicesKey(KEY_ALL_ACCESS, &hKeyDevices);
	if (fKeyRet != ERROR_SUCCESS)
		bRet = FALSE;
	else
	{
		// Delete device configuration key
		fKeyRet = RegDeleteKey(hKeyDevices, pszName);

		bRet = (fKeyRet==ERROR_SUCCESS);
		RegCloseKey(hKeyDevices);
	}

	return bRet;
}

/////////////////// HwEnumerateDevices()

BOOL _stdcall HwEnumerateDevices(DWORD nIndex, LPSTR pszName, HWDEVICE_CONFIG* pConfig)
{
	// pConfig could be a smaller pre 3.3 structure (for DLL builds)
	HWDEVICE_CONFIG config;
	memset(&config, 0, sizeof(HWDEVICE_CONFIG));
	config.cbSize = sizeof(HWDEVICE_CONFIG);

	// Call new ...Ex routine
	BOOL bRet = HwEnumerateDevicesEx(nIndex, pszName, &config);

	// Copy from new structure to old
	memcpy(pConfig, &config, HWDEVICE_CONFIG_PRE33_SIZE);
	pConfig->bNonIsaInterface = (config.fInterfaceType != DXISA_INTERFACE);
	return bRet;
}

/////////////////// HwEnumerateDevicesEx()

BOOL _stdcall HwEnumerateDevicesEx(DWORD nIndex, LPSTR pszName, HWDEVICE_CONFIG* pConfig)
{
	HKEY 		hKeyUserDevices;
	HKEY		hKeyDevice;
	DWORD		cchBufSize=MAX_NAME_LEN;
	FILETIME	time;
	LONG		fRet;

	if (OpenUserDevicesKey(KEY_READ, &hKeyUserDevices) != ERROR_SUCCESS)
		return FALSE;

	// Get the device name at index position
	fRet = RegEnumKeyEx(hKeyUserDevices, nIndex, 
				pszName, &cchBufSize,
				NULL, NULL, NULL, &time);

	if (fRet != ERROR_SUCCESS)
	{
		RegCloseKey(hKeyUserDevices);
		return FALSE;
	}

	// Open the device key
	fRet = RegOpenKeyEx(hKeyUserDevices, 
				pszName, 0, KEY_READ, &hKeyDevice);

    if (fRet != ERROR_SUCCESS)
    {
        RegCloseKey(hKeyUserDevices);
        return FALSE;
    }

	// Get the device's configuration information
	GetDeviceConfig(hKeyDevice, pConfig);

	RegCloseKey(hKeyDevice);
	RegCloseKey(hKeyUserDevices);
	return TRUE;
}

BOOL _stdcall HwGetDeviceConfig(LPCSTR pszName, HWDEVICE_CONFIG* pConfig)
{
	// Could be an old pre3.3 structure. Copy to new structure type.
	HWDEVICE_CONFIG config;
	memset(&config, 0, sizeof(HWDEVICE_CONFIG));
	config.cbSize = sizeof(HWDEVICE_CONFIG);

	BOOL bRet = HwGetDeviceConfigEx(pszName, &config);
	ASSERT(pConfig);

	// Copy 
	memcpy(pConfig, &config, HWDEVICE_CONFIG_PRE33_SIZE);
	return bRet;
}

/////////////////// HwGetDeviceConfigEx()
//
// Retrieves the registry configuration for a given user-defined device.
//
// Returns Value:
//
//		Returns TRUE if device key is found, else returns FALSE.

BOOL _stdcall HwGetDeviceConfigEx(LPCSTR pszName, HWDEVICE_CONFIG* pConfig)
{
	HKEY 	hKeyUserDevices;
	HKEY	hKeyDevice;
	LONG	fRet;

	if (OpenUserDevicesKey(KEY_READ, &hKeyUserDevices) != ERROR_SUCCESS)
		return FALSE;

	// Open the device key
	fRet = RegOpenKeyEx(hKeyUserDevices, 
				pszName, 0, KEY_READ, &hKeyDevice);

	if (fRet == ERROR_SUCCESS)
	{
		GetDeviceConfig(hKeyDevice, pConfig);
		RegCloseKey(hKeyDevice);
	}

	RegCloseKey(hKeyUserDevices);
	return (fRet == ERROR_SUCCESS);
}

/////////////////// HwZeroDeviceConfig()

VOID _stdcall HwZeroDeviceConfig(HWDEVICE_CONFIG* pConfig)
{
	ASSERT(pConfig);
	memset(pConfig, 0, HWDEVICE_CONFIG_PRE33_SIZE);
}

/////////////////// HwZeroDeviceConfigEx()

VOID _stdcall HwZeroDeviceConfigEx(HWDEVICE_CONFIG* pConfig)
{
	ASSERT(pConfig);
	memset(pConfig, 0, sizeof(HWDEVICE_CONFIG));
    pConfig->cbSize = sizeof(HWDEVICE_CONFIG);
	pConfig->fInterfaceType = DXISA_INTERFACE;
}

VOID _stdcall HwConfigureInterface(HWDEVICE_CONFIG* pConfig, 
	DWORD fInterfaceType, DWORD nBusNumber)
{
	ASSERT(pConfig);
	pConfig->bNonIsaInterface = (fInterfaceType != 1);
	pConfig->fInterfaceType = fInterfaceType;
	pConfig->nBusNumber = nBusNumber;
}

/////////////////// HwConfigureInterrupt()
//
// Stores an interrupt configuration in a HWDEVICE_CONFIG structure for
// a later call to HwConfigureDevice.

VOID _stdcall HwConfigureInterrupt(HWDEVICE_CONFIG* pConfig, DWORD nLevel)
{
	ASSERT(pConfig);
	pConfig->nInterruptLevel = nLevel;
	pConfig->bUsesInterrupt = TRUE;
}

/////////////////// HwConfigureIsrOp()

VOID _stdcall HwConfigureIsrOp(HWDEVICE_CONFIG* pConfig, DWORD nIsrOp,
		DWORD nIsrOffset, DWORD nIsrData)
{
	ASSERT(pConfig);
	pConfig->nIsrOp = nIsrOp;
	pConfig->nIsrOffset = nIsrOffset;
	pConfig->nIsrData = nIsrData;
}

VOID _stdcall HwConfigureIsrIdOp(HWDEVICE_CONFIG* pConfig, DWORD nOp, 
				DWORD nOffset, DWORD nMask, DWORD nValue)
{
	ASSERT(pConfig);
	pConfig->nIsrIdOp		= nOp;
	pConfig->nIsrIdOffset	= nOffset;
	pConfig->nIsrIdMask		= nMask;
	pConfig->nIsrIdValue	= nValue;
}

/////////////////// HwConfigurePortRange()
//
// Defines a port range in a HWDEVICE_CONFIG structure for a call to
// HwConfigureDevice.

VOID _stdcall HwConfigurePortRange(HWDEVICE_CONFIG* pConfig,
				DWORD nPortBase, DWORD nPortCount)
{
	ASSERT(pConfig);

	pConfig->nPortBase = nPortBase;
	pConfig->nPortCount = nPortCount;
	pConfig->bUsesPortRange = TRUE;
}

/////////////////// HwConfigureIgnoreConflicts()

VOID _stdcall HwConfigureIgnoreConflicts(HWDEVICE_CONFIG* pConfig, BOOL bEnable)
{
	pConfig->bIgnoreConflicts = bEnable;
}

/////////////////// HwConfigureMemRange()
//
// Defines a memory range in a HWDEVICE_CONFIG structure for a call to
// HwConfigureDevice.

VOID _stdcall HwConfigureMemRange(HWDEVICE_CONFIG* pConfig,
				DWORD nMemBase, DWORD nMemCount)
{
	ASSERT(pConfig);

	pConfig->nMemBase = nMemBase;
	pConfig->nMemCount = nMemCount;
	pConfig->bUsesMemRange = TRUE;
}

/////////////////// HwConfigureMemoryIo()

VOID _stdcall HwConfigureMemoryIo(HWDEVICE_CONFIG* pConfig,
			BOOL bMemoryIo)
{
	ASSERT(pConfig);
	pConfig->bMemoryIo = bMemoryIo;
}

/////////////////// HwIsDriverConfigured()
//
// Returns TRUE if there is a registry entry for the given driver name.

BOOL _stdcall HwIsDriverConfigured()
{
	HKEY hKeyWnsDrvr;

	if (OpenDriverKey(_T("DriverX"), KEY_READ, &hKeyWnsDrvr) == ERROR_SUCCESS)
	{
		RegCloseKey(hKeyWnsDrvr);
		return TRUE;
	}
	else
		return FALSE;
}

/////////////////// QueryDwordValue()

static LONG QueryDwordValue(HKEY hKey, LPTSTR pszValue, PDWORD pdwValue)
{
	DWORD dwSize = sizeof(DWORD);
	DWORD dwType;

	ASSERT(pszValue);
	ASSERT(pdwValue);

	return RegQueryValueEx(hKey, pszValue, NULL, &dwType,
		(unsigned char*) pdwValue, &dwSize);
}

/////////////////// SetDwordValue()

static LONG SetDwordValue(HKEY hKey, LPCTSTR pszValue, DWORD dwValue)
{
	return RegSetValueEx(hKey, pszValue,
				0, REG_DWORD, 
				(CONST BYTE*) &dwValue,
				sizeof(DWORD));
}

/////////////////// GetDeviceConfig()
//
// Helper routine to retrieve the registry configuration for a given 
// user-defined device.
// 
// Parameters:
//
//		hKeyDevice (IN) - handle to device key.
//		pConfig (OUT) - returns configuration of device.

static VOID GetDeviceConfig(HKEY hKeyDevice, HWDEVICE_CONFIG* pConfig)
{
	LONG	fRet;
	DWORD	dwValue;

	HwZeroDeviceConfigEx(pConfig);

	fRet = QueryDwordValue(hKeyDevice, _T("BusNumber"), &dwValue);
	if (fRet == ERROR_SUCCESS)
		pConfig->nBusNumber = dwValue;

	// See if interface is defined
	fRet = QueryDwordValue(hKeyDevice, _T("InterfaceType"), &dwValue);
	if (fRet == ERROR_SUCCESS)
	{
		pConfig->fInterfaceType = dwValue;

		if (dwValue == 5 /* PCI bus */)
		{
			fRet = QueryDwordValue(hKeyDevice, _T("VendorID"), &dwValue);
			if (fRet == ERROR_SUCCESS)
				pConfig->nVendorID = dwValue;

			fRet = QueryDwordValue(hKeyDevice, _T("DeviceID"), &dwValue);
			if (fRet == ERROR_SUCCESS)
				pConfig->nDeviceID = dwValue;
		}
	}

	// See if PortBase and PortCount values are defined.
	fRet = QueryDwordValue(hKeyDevice, _T("PortBase"), &dwValue);
	if (fRet != ERROR_SUCCESS)
		pConfig->bUsesPortRange = FALSE;
	else
	{
		pConfig->bUsesPortRange = TRUE;
		pConfig->nPortBase = dwValue;

		fRet = QueryDwordValue(hKeyDevice, _T("PortCount"), &dwValue);
		if (fRet != ERROR_SUCCESS)
		{
			pConfig->bUsesPortRange = FALSE;
			pConfig->nPortBase = 0;
		}
		else
			pConfig->nPortCount = dwValue;
	}

	// See if an interrupt level is defined
	fRet = QueryDwordValue(hKeyDevice, _T("InterruptLevel"), &dwValue);
	if (fRet != ERROR_SUCCESS)
		pConfig->bUsesInterrupt = FALSE;
	else
	{	
		pConfig->bUsesInterrupt = TRUE;
		pConfig->nInterruptLevel = dwValue;
    }

	// Look for optional ISR op (added in 3.0)
	fRet = QueryDwordValue(hKeyDevice, _T("IsrOp"), &dwValue);
	if (fRet == ERROR_SUCCESS && dwValue)
	{
		pConfig->nIsrOp = dwValue;

		fRet = QueryDwordValue(hKeyDevice, _T("IsrOffset"), &dwValue);
		if (fRet == ERROR_SUCCESS)
			pConfig->nIsrOffset = dwValue;

		fRet = QueryDwordValue(hKeyDevice, _T("IsrData"), &dwValue);
		if (fRet == ERROR_SUCCESS)
			pConfig->nIsrData = dwValue;
	}

	// Look for optional ISR ID op (added in 3.3)
	fRet = QueryDwordValue(hKeyDevice, _T("IsrIdOp"), &dwValue);
	if (fRet == ERROR_SUCCESS && dwValue)
	{
		pConfig->nIsrIdOp = dwValue;

		fRet = QueryDwordValue(hKeyDevice, _T("IsrIdOffset"), &dwValue);
		if (fRet == ERROR_SUCCESS)
			pConfig->nIsrIdOffset = dwValue;

		fRet = QueryDwordValue(hKeyDevice, _T("IsrIdMask"), &dwValue);
		if (fRet == ERROR_SUCCESS)
			pConfig->nIsrIdMask = dwValue;

		fRet = QueryDwordValue(hKeyDevice, _T("IsrIdValue"), &dwValue);
		if (fRet == ERROR_SUCCESS)
			pConfig->nIsrIdValue = dwValue;
	}

	// See if device is dynamic (added in 2.8)
	fRet = QueryDwordValue(hKeyDevice, _T("Dynamic"), &dwValue);
	if (fRet != ERROR_SUCCESS)
		pConfig->bDynamic = FALSE;
	else
		pConfig->bDynamic = (dwValue ? TRUE : FALSE);

	// Added mem range in 2.8
	fRet = QueryDwordValue(hKeyDevice, _T("MemBase"), &dwValue);
	if (fRet != ERROR_SUCCESS)
		pConfig->bUsesMemRange = FALSE;
	else
	{
		pConfig->bUsesMemRange = TRUE;
		pConfig->nMemBase = dwValue;

		fRet = QueryDwordValue(hKeyDevice, _T("MemCount"), &dwValue);
		if (fRet != ERROR_SUCCESS)
		{
			pConfig->bUsesMemRange = FALSE;
			pConfig->nMemBase = 0;
		}
		else
			pConfig->nMemCount = dwValue;
	}

	// See if device ignores resource conflicts (added in 3.2)
	if (pConfig->fInterfaceType != DXPCI_INTERFACE)
	{
		fRet = QueryDwordValue(hKeyDevice, _T("IgnoreConflicts"), &dwValue);
		if (fRet != ERROR_SUCCESS)
			pConfig->bIgnoreConflicts = FALSE;
		else
			pConfig->bIgnoreConflicts = (dwValue ? TRUE : FALSE);
	}

	// Added MemoryIo flag in 3.3 for PCI devices with memory-mapped registers
	fRet = QueryDwordValue(hKeyDevice, _T("MemoryIo"), &dwValue);
	if (fRet != ERROR_SUCCESS)
		pConfig->bMemoryIo = FALSE;
	else
		pConfig->bMemoryIo = (dwValue ? TRUE : FALSE);

	if (pConfig->cbSize >= sizeof(HWDEVICE_CONFIG))
	{
		fRet = QueryDwordValue(hKeyDevice, _T("DmaBufSize"), &dwValue);
		if (fRet == ERROR_SUCCESS)
			pConfig->nDmaBufSize = dwValue;
	}
}

/////////////////// OpenKey()
//
// Helper function to open a registry key and optionally create it if
// not found.
//
// Returns ERROR_SUCCESS if successful, else an Win32 error code.

static LONG OpenKey(HKEY hParentKey, LPCSTR pszSubkeyName, REGSAM samDesired,
	BOOL bCreate, PHKEY phSubkey)
{
	LONG fRet;

	ASSERT(phSubkey);

	fRet = RegOpenKeyEx(hParentKey,
					pszSubkeyName,
					0, 
					samDesired, 
					phSubkey);

	if (fRet != ERROR_SUCCESS && bCreate)
	{
		DWORD dwDisposition;

		fRet = RegCreateKeyEx(hParentKey,
					pszSubkeyName,
					0,
					NULL,
					0,
					samDesired,
					NULL,
					phSubkey,
					&dwDisposition);
	}

	return fRet;
}

/////////////////// OpenUserDevicesKey()
//
// Returns the key under which user device registry keys. On Windows 95, 
// this is the DRIVERX key. On NT, it is the DriverX/Parameters key. Users
// should call RegCloseKey with the returned key value.
//
// Return Value:
//
//		ERROR_SUCCESS or a Win32 error code if function fails.

static LONG OpenUserDevicesKey(REGSAM samDesired, PHKEY phKeyUserDevices)
{
	HKEY hKeyDriver;
	LONG fRet;

	ASSERT(phKeyUserDevices);

	// Open driver key
	fRet = OpenDriverKey(_T("DriverX"), samDesired, &hKeyDriver);
	if (fRet == ERROR_SUCCESS)
	{
		if (IsOn95())
		{
			// On Windows 95, user devices are under DriverX key
			*phKeyUserDevices = hKeyDriver;
		}
		else
		{
			// On NT, user devices are under a Parameters subkey
			fRet = OpenKey(hKeyDriver, _T("Parameters"), KEY_ALL_ACCESS,
				TRUE, phKeyUserDevices);
			RegCloseKey(hKeyDriver);
		}
	}

	return fRet;
}

/////////////////// OpenDriverKey()
//
// Opens a driver registry key for the given driver name. Caller should
// call RegCloseKey with returned key handle if function if successful.
//
// Return Value:
//
//		ERROR_SUCCESS if successful, else a Win32 error code.

static LONG OpenDriverKey(LPCSTR pszName, REGSAM samDesired, PHKEY phKeyDriver)
{
	ASSERT(pszName);
	ASSERT(phKeyDriver);

	LONG fRet;
	char szFullPath[80];

	if (IsOn95())
	{
		_stprintf(szFullPath, 
			_T("System\\CurrentControlSet\\Services\\VxD\\%s"), pszName);

		fRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
			szFullPath,
			0, 
			samDesired, 
			phKeyDriver);
	}
	else
	{
		_stprintf(szFullPath, 
			_T("System\\CurrentControlSet\\Services\\%s"), pszName);

		fRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
			szFullPath,
			0, 
			samDesired, 
			phKeyDriver);
	}

	return fRet;
}

/////////////////// SetRegistryConfiguration()
//
// Helper routine to store configuration values under a user-defined
// device registry key.

static LONG SetRegistryConfiguration(HKEY hKeyDevice, 
	HWDEVICE_CONFIG* pConfig)
{
	LONG fRet = ERROR_SUCCESS;

	// Delete any existing configuration
	RegDeleteValue(hKeyDevice, _T("InterfaceType"));
	RegDeleteValue(hKeyDevice, _T("BusNumber"));
	RegDeleteValue(hKeyDevice, _T("VendorID"));
	RegDeleteValue(hKeyDevice, _T("DeviceID"));
	RegDeleteValue(hKeyDevice, _T("PortBase"));
	RegDeleteValue(hKeyDevice, _T("PortCount"));
	RegDeleteValue(hKeyDevice, _T("InterruptLevel"));
	RegDeleteValue(hKeyDevice, _T("IsrOp"));
	RegDeleteValue(hKeyDevice, _T("IsrData"));
	RegDeleteValue(hKeyDevice, _T("IsrOffset"));
	RegDeleteValue(hKeyDevice, _T("Dynamic"));
	RegDeleteValue(hKeyDevice, _T("IgnoreConflicts"));
	RegDeleteValue(hKeyDevice, _T("MemBase"));
	RegDeleteValue(hKeyDevice, _T("MemCount"));
	RegDeleteValue(hKeyDevice, _T("IsrIdOp"));
	RegDeleteValue(hKeyDevice, _T("IsrIdMask"));
	RegDeleteValue(hKeyDevice, _T("IsrIdOffset"));
	RegDeleteValue(hKeyDevice, _T("IsrIdValue"));
	RegDeleteValue(hKeyDevice, _T("MemoryIo"));
	RegDeleteValue(hKeyDevice, _T("DmaBufSize"));

	// Store interface
	if (pConfig->fInterfaceType != DXISA_INTERFACE)
	{
		fRet = SetDwordValue(hKeyDevice, _T("InterfaceType"), pConfig->fInterfaceType);
		if (fRet != ERROR_SUCCESS)
			return fRet;

		if (pConfig->fInterfaceType == DXPCI_INTERFACE)
		{
			// Set vendor and device ID values for PCI devices
			fRet = SetDwordValue(hKeyDevice, _T("VendorID"), pConfig->nVendorID);
			if (fRet != ERROR_SUCCESS)
				return fRet;

			fRet = SetDwordValue(hKeyDevice, _T("DeviceID"), pConfig->nDeviceID);
			if (fRet != ERROR_SUCCESS)
				return fRet;
		}
		else
		{
			fRet = SetDwordValue(hKeyDevice, _T("BusNumber"), pConfig->nBusNumber);
			if (fRet != ERROR_SUCCESS)
				return fRet;
		}
	}

	// Store port range
	if (pConfig->bUsesPortRange)
	{
		fRet = SetDwordValue(hKeyDevice, _T("PortBase"), pConfig->nPortBase);
		if (fRet != ERROR_SUCCESS)
			return fRet;
		fRet = SetDwordValue(hKeyDevice, _T("PortCount"), pConfig->nPortCount);
		if (fRet != ERROR_SUCCESS)
			return fRet;
	}

	// Store interrupt
	if (pConfig->bUsesInterrupt)
	{
		fRet = SetDwordValue(hKeyDevice, _T("InterruptLevel"), pConfig->nInterruptLevel);
		if (fRet != ERROR_SUCCESS)
			return fRet;
	}

	// Store ISR op -added in 3.0
	if (pConfig->nIsrOp)
	{
		fRet = SetDwordValue(hKeyDevice, _T("IsrOp"), pConfig->nIsrOp);
		if (fRet != ERROR_SUCCESS)
			return fRet;

		fRet = SetDwordValue(hKeyDevice, _T("IsrOffset"), pConfig->nIsrOffset);
		if (fRet != ERROR_SUCCESS)
			return fRet;

		fRet = SetDwordValue(hKeyDevice, _T("IsrData"), pConfig->nIsrData);
		if (fRet != ERROR_SUCCESS)
			return fRet;
	}

	// Store ISR ID op -added in 3.3
	if (pConfig->nIsrIdOp)
	{
		fRet = SetDwordValue(hKeyDevice, _T("IsrIdOp"), pConfig->nIsrIdOp);
		if (fRet != ERROR_SUCCESS)
			return fRet;

		fRet = SetDwordValue(hKeyDevice, _T("IsrIdOffset"), pConfig->nIsrIdOffset);
		if (fRet != ERROR_SUCCESS)
			return fRet;

		fRet = SetDwordValue(hKeyDevice, _T("IsrIdMask"), pConfig->nIsrIdMask);
		if (fRet != ERROR_SUCCESS)
			return fRet;

		fRet = SetDwordValue(hKeyDevice, _T("IsrIdValue"), pConfig->nIsrIdValue);
		if (fRet != ERROR_SUCCESS)
			return fRet;
	}

	// Store dynamic flag -added in 2.8
	if (pConfig->bDynamic)
	{
		fRet = SetDwordValue(hKeyDevice, _T("Dynamic"), 1);
		if (fRet != ERROR_SUCCESS)
			return fRet;
	}

	// Store memory range -added in 2.8
	if (pConfig->bUsesMemRange)
	{
		fRet = SetDwordValue(hKeyDevice, _T("MemBase"), pConfig->nMemBase);
		if (fRet != ERROR_SUCCESS)
			return fRet;

		fRet = SetDwordValue(hKeyDevice, _T("MemCount"), pConfig->nMemCount);
		if (fRet != ERROR_SUCCESS)
			return fRet;
	}

	// nVendorID doubles as 'ignore conflicts' flag in 3.2
	if (pConfig->fInterfaceType != DXPCI_INTERFACE)
	{
		if (pConfig->bIgnoreConflicts)
		{
			fRet = SetDwordValue(hKeyDevice, _T("IgnoreConflicts"), 1);
			if (fRet != ERROR_SUCCESS)
				return fRet;
		}
	}

	if (pConfig->bMemoryIo)
	{
		fRet = SetDwordValue(hKeyDevice, _T("MemoryIo"), 1);
		if (fRet != ERROR_SUCCESS)
			return fRet;
	}

	if (pConfig->cbSize >= sizeof(HWDEVICE_CONFIG))
	{
		fRet = SetDwordValue(hKeyDevice, _T("DmaBufSize"), pConfig->nDmaBufSize);
		if (fRet != ERROR_SUCCESS)
			return fRet;
	}

	RegFlushKey(hKeyDevice);
	return(fRet);
}

/////////////////// HwStartDevice()
//
// This routine (added in 2.8) replaces HwStartDriver and should be
// used in it's place to start an individual device without disrupting
// other devices.

BOOL _stdcall HwStartDevice(LPCSTR pszName)
{
	ASSERT(pszName);
	HANDLE hDevice;

	if (HwIsDeviceRunning(pszName))
		return TRUE;

	// Get handle to VxD or _Default device
	if (IsOn95())
	{
		hDevice = CreateFile(_T("\\\\.\\DRIVERX.VXD"), 
			0, 
			0,
			NULL, 
			0, 
			0,
			NULL);
	}
	else
	{
		hDevice = CreateFile(_T("\\\\.\\_Default30"),
						GENERIC_READ|GENERIC_WRITE,
						0,
						NULL,
						OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL,
						NULL); 
	}

	if (hDevice == INVALID_HANDLE_VALUE)
	{
		TRACE(_T("DriverX: Unable to open handle in HwStartDevice.\n"));
		return FALSE;
	}

	BOOL bRet;
	DWORD cbRet;

	// Send command to start device
	bRet = DeviceIoControl(hDevice,
				(DWORD)IOCTL_START_DEVICE,
				(PVOID)pszName,
				strlen(pszName)+1, // Always an ANSI string
				NULL,
				0,
				&cbRet,
				NULL);

	if (!bRet)
		TRACE(_T("DriverX: Failed to start device %s.\n"), pszName);

	CloseHandle(hDevice);
	return bRet;
}

/////////////////// HwStopDevice()
//
// This routine (added in 2.8) replaces HwStopDevice and should be
// used in it's place to stop an individual device without disrupting
// other devices.

BOOL _stdcall HwStopDevice(LPCSTR pszName)
{
	ASSERT(pszName);
	HANDLE hDevice;

	// Get handle to VxD or _Default device
	if (IsOn95())
	{
		hDevice = CreateFile(_T("\\\\.\\DRIVERX.VXD"), 
			0, 
			0,
			NULL, 
			0, 
			0,
			NULL);
	}
	else
	{
		hDevice = CreateFile(_T("\\\\.\\_Default30"),
						GENERIC_READ|GENERIC_WRITE,
						0,
						NULL,
						OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL,
						NULL); 
	}

	if (hDevice == INVALID_HANDLE_VALUE)
	{
		TRACE(_T("DriverX: Unable to open _Default30 device.\n"));
		return FALSE;
	}

	// Send command to stop device

	DWORD cbRet;
	char szName[MAX_DEVICE_NAME_LEN+1];

	if (IsOnNT())
		sprintf(szName, "\\Device\\%s", pszName);
	else
		strcpy(szName, pszName);

	DeviceIoControl(hDevice,
				(DWORD)IOCTL_STOP_DEVICE,
				(PVOID)szName,
				strlen(szName)+1, // Always an ANSI string
				NULL,
				0,
				&cbRet,
				NULL);

	CloseHandle(hDevice);
	return TRUE;
}

/////////////////// HwIsDeviceRunning()
//
// Returns TRUE if a device object is currently running and ready
// for use.

BOOL _stdcall HwIsDeviceRunning(LPCTSTR pszDevice)
{
	HWDEVICE* pDevice = HwNewDevice(NULL);
	BOOL bRet = HwConnectDevice(pDevice, pszDevice);
	HwDeleteDevice(pDevice);
	return bRet;
}

/////////////////// HwMapPorts()

BOOL _stdcall HwMapPorts(HWDEVICE* pDevice, DWORD nBase, DWORD nCount, PDWORD pnIndex)
{
	return HwMapPortsEx(pDevice, nBase, nCount, 1 /* Isa */, 0, pnIndex);
}

/////////////////// HwMapPortsEx()
//
// Sets up a port range and returns an index value for identifying the
// port range in other calls.
//
// Parameters:
//
//		pnIndex (IN/OUT) -desired port range index to configure or (DWORD)(-1)
//			to pick an open range. Returns mapped port range index for
//			all cases.
//
// Return Value:
//
//		Returns TRUE if successful, FALSE on error.

BOOL _stdcall HwMapPortsEx(HWDEVICE* pDevice, DWORD nBase, DWORD nCount, 
		DWORD fInterfaceType, DWORD nBusNumber, PDWORD pnIndex)
{
	BOOL bRet;
	MAP_PORTS_OUTPUT mapout;
	MAP_PORTS_INPUT mapin;
	DWORD cbReturned=0;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(pnIndex);
	
	memset(&mapin, 0, sizeof(MAP_PORTS_INPUT));

	mapin.nDeviceID		 = pDevice->nID;
	mapin.fInterfaceType = fInterfaceType;
	mapin.nBusNumber	 = nBusNumber;
	mapin.nPortBase		 = nBase;
	mapin.nPortCount	 = nCount;
	mapin.nIndex		 = *pnIndex;

	bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_MAP_PORTS,
					&mapin,
					sizeof(MAP_PORTS_INPUT),
					&mapout,
					sizeof(MAP_PORTS_OUTPUT),
					&cbReturned,
					NULL );

	if (!bRet || (cbReturned!=sizeof(MAP_PORTS_OUTPUT)))
	{
		TRACE(_T("DriverX: IOCTL_MAP_PORTS failed.\n"));	
		bRet = FALSE;
	}
	else
	{
		if (!mapout.nHalReturn)
		{
			TRACE(_T("DriverX: mapping of port range failed.\n"));
			bRet = FALSE;
		}
		else
			*pnIndex = mapout.nPortIndex;
	}

	return bRet;
}

/////////////////// HwUnmapPorts()
//
// Unmaps a port range.

BOOL _stdcall HwUnmapPorts(HWDEVICE* pDevice, DWORD nIndex)
{
	UNMAP_PORTS_INPUT unmapin;
	DWORD cbReturned;
	BOOL bRet;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);

	memset(&unmapin, 0, sizeof(UNMAP_PORTS_INPUT));
	unmapin.nDeviceID = pDevice->nID;
	unmapin.nPortIndex = nIndex;

	bRet = DeviceIoControl(pDevice->hDevice,
				(DWORD)IOCTL_UNMAP_PORTS,
				&unmapin,
				sizeof(UNMAP_PORTS_INPUT),
				NULL,
				0,
				&cbReturned,
				NULL);

	if (!bRet)
		TRACE(_T("DriverX: IOCTL_UNMAP_PORTS failed!\n"));	

	return bRet;
}

/////////////////// HwInPort()
//
// Executes a single port read operation.
//
// Parameters:
//
//		pDevice (IN) - target device.
//		nIndex (IN) - specifies target port range.
//		nOffset (IN) - offset into port range.
//		nDataSize (IN) - data size (1-byte, 2-word, 4-long).
//
// Return Value:
//
//		Port value.

DWORD _stdcall HwInPort(HWDEVICE* pDevice, DWORD nIndex, DWORD nOffset, DWORD nDataSize)
{
	// Setup an inp operation
	DXOPHEADER header;
	InitializeOpHeader(&header, nIndex, 1, 0);
	HwQueueInPort(&header, nOffset, nDataSize, NULL);

	// Execute the operation (HwExecute calls handler on errors)
	HwExecute(pDevice, &header);
	return header.ops[0].Inp.nValue;
}

// Simple routines targeting port range 0

BYTE _stdcall HwInp(HWDEVICE* pDevice, DWORD nOffset)
{ 
	return (BYTE) HwInPort(pDevice, 0, nOffset, sizeof(BYTE)); 
}

WORD _stdcall HwInpw(HWDEVICE* pDevice, DWORD nOffset)
{ 
	return (WORD) HwInPort(pDevice, 0, nOffset, sizeof(WORD)); 
}

DWORD _stdcall HwInpl(HWDEVICE* pDevice, DWORD nOffset)
{ 
	return HwInPort(pDevice, 0, nOffset, sizeof(DWORD)); 
}

/////////////////// HwOutPort()
//
// Executes a single port write operation.
//
// Parameters:
//
//		pDevice (IN) - target device.
//		nIndex (IN) - specifies target port range.
//		nOffset (IN) - offset into port range.
//		nDataSize (IN) - data size (1-byte, 2-word, or 4-long)
//		nData (IN) - data to write
//
// Return Value:
//
//		Return from HwExecute (TRUE if successful, FALSE on error)

BOOL _stdcall HwOutPort(HWDEVICE* pDevice, DWORD nIndex, DWORD nOffset, 
		DWORD nDataSize, DWORD nData)
{
	// Setup an inp operation
	DXOPHEADER header;
	InitializeOpHeader(&header, nIndex, 1, 0);
	HwQueueOutPort(&header, nOffset, nDataSize, nData);

	// Execute the operation (HwExecute calls handler on errors)
	return HwExecute(pDevice, &header);
}

// Simple routines targeting port range 0

BOOL _stdcall HwOutp(HWDEVICE* pDevice, DWORD nOffset, DWORD nData)
{
	return HwOutPort(pDevice, 0, nOffset, sizeof(BYTE), nData);
}

BOOL _stdcall HwOutpw(HWDEVICE* pDevice, DWORD nOffset, DWORD nData)
{
	return HwOutPort(pDevice, 0, nOffset, sizeof(WORD), nData);
}

BOOL _stdcall HwOutpl(HWDEVICE* pDevice, DWORD nOffset, DWORD nData)
{
	return HwOutPort(pDevice, 0, nOffset, sizeof(DWORD), nData);
}

/////////////////////////////////////////////////////////////////////////////
// HwAutoIncReadPort()

BOOL _stdcall HwAutoIncReadPort(HWDEVICE* pDevice, DWORD nIndex, DWORD nOffset, 
		PVOID pBuf, DWORD nCount, DWORD nDataSize)
{
	AUTOINC_RW_INPUT input;
	BOOL bRet;
	DWORD cbReturned;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(pBuf);

	memset(&input, 0, sizeof(AUTOINC_RW_INPUT));

	input.nDeviceID	= pDevice->nID;
	input.nIndex	= nIndex;
	input.nOffset	= nOffset;
	input.nCount	= nCount;
	input.nDataSize	= nDataSize;

	// DriverX ActiveX control reads bytes into short array
	if (nDataSize == BYTES_INTO_SHORT_ARRAY) 
	{
		input.nDataSize   = 1;
		input.bShortArray = TRUE;
	}

	if ((nCount > USE_DIRECT) || IsOn95())
	{
		bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_AUTOINC_READ,
					&input,
					sizeof(AUTOINC_RW_INPUT),
					pBuf,
					nCount,
					&cbReturned,
					NULL);

		if (!bRet)
			TRACE(_T("DriverX: Autoinc read DeviceIoControl failed!\n"));		
	}
	else
	{
		bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_BUF_AUTOINC_READ,
					&input,
					sizeof(AUTOINC_RW_INPUT),
					pBuf,
					nCount,
					&cbReturned,
					NULL );

		if (!bRet)
			TRACE(_T("DriverX: Buffered autoinc read DeviceIoControl failed!\n"));		
	}

	return bRet;
}

/////////////////// HwAutoIncWritePort()

BOOL _stdcall HwAutoIncWritePort(HWDEVICE* pDevice, DWORD nIndex, DWORD nOffset, PVOID pBuf,
	DWORD nCount, DWORD nDataSize)
{
	AUTOINC_RW_INPUT input;
	BOOL bRet;
	// note: buffer only used if data size > USE_DIRECT
	unsigned char buf[sizeof(AUTOINC_RW_INPUT)+USE_DIRECT];
	DWORD cbReturned;

	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(pBuf);

	memset(&input, 0, sizeof(AUTOINC_RW_INPUT));

 	input.nDeviceID	= pDevice->nID;
	input.nIndex	= nIndex;
	input.nOffset	= nOffset;
	input.nCount	= nCount;
	input.nDataSize	= nDataSize;

	// DriverX ActiveX control reads bytes into short array
	if (nDataSize == BYTES_INTO_SHORT_ARRAY) 
	{
		input.nDataSize   = 1;
		input.bShortArray = TRUE;
	}

	if ((nCount > USE_DIRECT) || IsOn95())
	{
		bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_AUTOINC_WRITE,
					&input,
					sizeof(AUTOINC_RW_INPUT),
					pBuf,
					nCount,
					&cbReturned,
					NULL );

		if (!bRet)
			TRACE(_T("DriverX: Autoinc write DeviceIoControl failed!\n"));		
	}
	else
	{
		// For buffered mode, append data to end of input structure
		memcpy(buf, &input, sizeof(AUTOINC_RW_INPUT));
		memcpy(&buf[sizeof(AUTOINC_RW_INPUT)], pBuf, nCount);

		bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD)IOCTL_BUF_AUTOINC_WRITE,
					buf,
					sizeof(AUTOINC_RW_INPUT) + nCount,
					NULL,
					0,
					&cbReturned,
					NULL );

		if (!bRet)
			TRACE(_T("DriverX: Buffered autoinc write DeviceIoControl failed!\n"));
	}

	return bRet;
}

/////////////////// HwGetBusDataByOffset()

BOOL _stdcall HwGetBusDataByOffset(HWDEVICE* pDevice, ULONG nOffset, PVOID pBuffer, ULONG cbBuffer)
{
	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(pBuffer);

	BUS_DATA_PARAMS params;
	memset(&params, 0, sizeof(BUS_DATA_PARAMS));
	params.nDeviceID = pDevice->nID;
	params.nOffset   = nOffset;
	params.cbBuffer  = cbBuffer;

	DWORD cbReturned;
	BOOL bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD) IOCTL_GET_BUS_DATA_BY_OFFSET,
					&params,
					sizeof(params),
					pBuffer,
					cbBuffer,
					&cbReturned,
					NULL);

	if (!bRet)
		TRACE(_T("DriverX: IOCTL_GET_BUS_DATA_BY_OFFSET failed.\n"));

	return bRet;
}

/////////////////// HwSetBusDataByOffset()

BOOL _stdcall HwSetBusDataByOffset(HWDEVICE* pDevice, ULONG nOffset, PVOID pBuffer, ULONG cbBuffer)
{
	ASSERT(pDevice);
	ASSERT(pDevice->hDevice != INVALID_HANDLE_VALUE);
	ASSERT(pBuffer);

	ULONG nSize = sizeof(BUS_DATA_PARAMS) + cbBuffer;
	PBUS_DATA_PARAMS pParams = (PBUS_DATA_PARAMS)(new char[nSize]);
	memset(pParams, 0, nSize);

	pParams->nDeviceID = pDevice->nID;
	pParams->nOffset   = nOffset;
	pParams->cbBuffer  = cbBuffer;
	memcpy(pParams->buffer, pBuffer, cbBuffer);

	DWORD cbReturned;
	BOOL bRet = DeviceIoControl(pDevice->hDevice,
					(DWORD) IOCTL_SET_BUS_DATA_BY_OFFSET,
					pParams,
					nSize,
					NULL,
					0,
					&cbReturned,
					NULL);

	delete [] pParams;
	if (!bRet)
		TRACE(_T("DriverX: IOCTL_SET_BUS_DATA_BY_OFFSET failed.\n"));

	return bRet;
}

			//////////////////////////////
			// C++ Class Implementations
			//////////////////////////////

/////////////////// static CDevice::IsrCallback()

VOID CDevice::IsrCallback(HWDEVICE* pHwdevice)
{
	ASSERT(pHwdevice);
	CDevice* pDevice = (CDevice*)(pHwdevice->pExtension);
	ASSERT(pDevice);
	pDevice->OnInterrupt();
}

/////////////////// CDevice::OnInterrupt()

VOID CDevice::OnInterrupt()
{
}

/////////////////// static CDevice::ErrorHandler()

VOID CDevice::ErrorHandler(HWDEVICE*, DWORD nError)
{
	switch (nError)
	{
	case HW_EVENT_ERROR:
		throw CDxException(nError, _T("Unable to allocate event."));
	case HW_WAIT_INTR_ERROR:
		throw CDxException(nError, _T("Error while waiting for interrupt."));
	case HW_IOCTL_ERROR:
		throw CDxException(nError, _T("DeviceIoControl call failed."));
	default:
		throw CDxException(nError, _T("No description."));
	}
}
